use std::{error::Error, sync::Arc};

use vulkano::{
    Validated,
    buffer::{AllocateBufferError, Buffer, BufferContents, BufferCreateInfo, Subbuffer},
    memory::allocator::{AllocationCreateInfo, StandardMemoryAllocator},
};

pub trait AsBindableBuffer {
    type BufferData: BufferContents + Clone;

    fn buffer_create_info() -> BufferCreateInfo;

    fn allocation_create_info() -> AllocationCreateInfo;

    fn to_buffer_data(&self) -> Self::BufferData;

    fn create_buffer(
        &self,
        memory_allocator: &Arc<StandardMemoryAllocator>,
    ) -> Result<Subbuffer<[Self::BufferData]>, Validated<AllocateBufferError>> {
        Buffer::from_iter(
            memory_allocator.clone(),
            Self::buffer_create_info(),
            Self::allocation_create_info(),
            [self.to_buffer_data()],
        )
    }

    fn update_buffer(&self, buffer: &Subbuffer<[Self::BufferData]>) -> Result<(), Box<dyn Error>> {
        let mut write_guard = buffer.write()?;
        write_guard[0] = self.to_buffer_data();
        Ok(())
    }
}

pub trait AsUniformBuffer: AsBindableBuffer {
    fn create_uniform_buffer(
        &self,
        memory_allocator: &Arc<StandardMemoryAllocator>,
    ) -> Result<Subbuffer<[Self::BufferData]>, Validated<AllocateBufferError>> {
        self.create_buffer(memory_allocator)
    }
}

pub trait AsVertexBuffer: AsBindableBuffer
where
    Self: Sized,
{
    fn create_vertex_buffer(
        memory_allocator: &Arc<StandardMemoryAllocator>,
        vertices: &[Self],
    ) -> Result<Subbuffer<[Self::BufferData]>, Validated<AllocateBufferError>> {
        let buffer_data: Vec<Self::BufferData> =
            vertices.iter().map(|v| v.to_buffer_data()).collect();

        Buffer::from_iter(
            memory_allocator.clone(),
            Self::buffer_create_info(),
            Self::allocation_create_info(),
            buffer_data,
        )
    }
}

pub trait AsIndexBuffer: AsBindableBuffer
where
    Self: Sized,
{
    fn create_index_buffer(
        memory_allocator: &Arc<StandardMemoryAllocator>,
        indices: &[Self],
    ) -> Result<Subbuffer<[Self::BufferData]>, Validated<AllocateBufferError>> {
        let buffer_data: Vec<Self::BufferData> =
            indices.iter().map(|i| i.to_buffer_data()).collect();

        Buffer::from_iter(
            memory_allocator.clone(),
            Self::buffer_create_info(),
            Self::allocation_create_info(),
            buffer_data,
        )
    }
}
